"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var node_sqlparser_1 = require("node-sqlparser");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var firestore_1 = require("rxfire/firestore");
var firesql_1 = require("../firesql");
var select_1 = require("../select");
var utils_1 = require("../utils");
firesql_1.FireSQL.prototype.rxQuery = function (sql) {
    utils_1.assert(typeof sql === 'string' && sql.length > 0, 'rxQuery() expects a non-empty string.');
    var ast = node_sqlparser_1.parse(sql);
    utils_1.assert(ast.type === 'select', 'Only SELECT statements are supported.');
    return rxSelect(this._getRef(), ast);
};
function rxSelect(ref, ast) {
    var queries = select_1.generateQueries(ref, ast);
    if (ast._next) {
        utils_1.assert(ast._next.type === 'select', ' UNION statements are only supported between SELECTs.');
        // This is the UNION of 2 SELECTs, so lets process the second
        // one and merge their queries
        queries = queries.concat(select_1.generateQueries(ref, ast._next));
        // FIXME: The SQL parser incorrectly attributes ORDER BY to the second
        // SELECT only, instead of to the whole UNION. Find a workaround.
    }
    // We need to determine if we have to include
    // the document's key (__name__) in the results.
    var includeKey = false;
    if (Array.isArray(ast.columns)) {
        for (var i = 0; i < ast.columns.length; i++) {
            if (ast.columns[i].expr.type === 'column_ref') {
                if (ast.columns[i].expr.column === '__name__') {
                    includeKey = true;
                    break;
                }
            }
        }
    }
    var rxData = rxjs_1.combineLatest(queries.map(function (query) {
        return firestore_1.collectionData(query, includeKey ? '__name__' : void 0);
    }));
    return rxData.pipe(operators_1.map(function (results) {
        // We have an array of results (one for each query we generated) where
        // each element is an array of documents. We need to flatten them.
        return results.reduce(function (docs, current) { return docs.concat(current); }, []);
    }), operators_1.map(function (documents) {
        return select_1.processDocuments(ast, queries, documents);
    }));
}
//# sourceMappingURL=index.js.map